"""
Exception classes.
"""


class ConfigurationError(Exception):
    """
    To be raised when the client developer has made an error in configuring their
    station or any of its relevant components.
    """
